@@ -31,7 +31,7 @@ extern zend_module_entry pinba_module_entry;
 
 #define PINBA_COLLECTOR_DEFAULT_PORT "30002"
 #define PINBA_COLLECTORS_MAX 8
-#define PHP_PINBA_VERSION "1.1.1"
+#define PHP_PINBA_VERSION "1.1.2"
 
 typedef struct _pinba_req_data { /* {{{ */
 	char *server_name;
@@ -47,12 +47,16 @@ typedef struct _pinba_req_data { /* {{{ */
 } pinba_req_data;
 /* }}} */
 
-typedef struct _pinba_collector {
+typedef struct {
+	int fd;
 	struct sockaddr_storage sockaddr;
 	size_t                  sockaddr_len; /* shouldn't this be socken_t ? */
+	time_t                  sockaddr_time; /* time last resolved */
+} pinba_sockaddr;
+
+typedef struct _pinba_collector {
 	char *host;
 	char *port;
-	int fd;
 } pinba_collector;
 
 ZEND_BEGIN_MODULE_GLOBALS(pinba) /* {{{ */
@@ -74,6 +78,7 @@ ZEND_BEGIN_MODULE_GLOBALS(pinba) /* {{{ */
 	zend_bool in_rshutdown;
 	zend_bool enabled;
 	zend_bool auto_flush;
+	time_t resolve_interval; /* seconds */
 ZEND_END_MODULE_GLOBALS(pinba)
 /* }}} */
 
