@@ -93,6 +93,8 @@ size_t (*old_sapi_ub_write) (const char *, size_t);
 #define PINBA_ONLY_RUNNING_TIMERS (1<<2)
 #define PINBA_AUTO_FLUSH (1<<3)
 
+static HashTable resolver_cache;
+
 typedef struct _pinba_timer_tag { /* {{{ */
 	char *name;
 	int name_len;
@@ -191,10 +193,6 @@ static inline void php_pinba_cleanup_collectors(pinba_collector *collectors, uns
 	for (i = 0; i < *n_collectors; i++) {
 		pinba_collector *collector = &collectors[i];
 
-		if (collector->fd >= 0) {
-			close(collector->fd);
-		}
-
 		if (collector->host) {
 			free(collector->host);
 		}
@@ -465,14 +463,73 @@ static size_t sapi_ub_write_counter(const char *str, size_t length) /* {{{ */
 }
 /* }}} */
 
-static int php_pinba_init_socket(pinba_collector *collectors, int n_collectors) /* {{{ */
+static pinba_sockaddr *php_pinba_connect_socket(const char *host, const char *port) /* {{{ */
 {
+	pinba_sockaddr *sa;
+	time_t now = time(NULL);
+	char *hostport;
+	size_t hostport_len;
+
+	hostport_len = spprintf(&hostport, 0, "%s:%s", host, port);
+
+	sa = zend_hash_str_find_ptr(&resolver_cache, hostport, hostport_len);
+	if (sa) {
+		if ((now - sa->sockaddr_time) < PINBA_G(resolve_interval)) {
+			efree(hostport);
+			return sa;
+		}
+		/* delete the value and proceed to re-resolve */
+		zend_hash_str_del(&resolver_cache, hostport, hostport_len);
+		sa = NULL;
+	}
+
 	struct addrinfo *ai_list;
 	struct addrinfo *ai_ptr = NULL;
 	struct addrinfo  ai_hints;
+
+	memset(&ai_hints, 0, sizeof(ai_hints));
+	ai_hints.ai_flags     = 0;
+#ifdef AI_ADDRCONFIG
+	ai_hints.ai_flags    |= AI_ADDRCONFIG;
+#endif
+	ai_hints.ai_family    = AF_UNSPEC;
+	ai_hints.ai_socktype  = SOCK_DGRAM;
+	ai_hints.ai_addr      = NULL;
+	ai_hints.ai_canonname = NULL;
+	ai_hints.ai_next      = NULL;
+
+	ai_list = NULL;
+	int status = getaddrinfo(host, port, &ai_hints, &ai_list);
+	if (status != 0) {
+		efree(hostport);
+		php_error_docref(NULL, E_WARNING, "failed to resolve Pinba server hostname '%s': %s", host, gai_strerror(status));
+		return NULL;
+	} else {
+		for (ai_ptr = ai_list; ai_ptr != NULL; ai_ptr = ai_ptr->ai_next) {
+			int fd = socket(ai_ptr->ai_family, ai_ptr->ai_socktype, ai_ptr->ai_protocol);
+			if (fd >= 0) {
+
+				sa = pecalloc(1, sizeof(pinba_sockaddr), 1);
+				memcpy(&sa->sockaddr, ai_ptr->ai_addr, ai_ptr->ai_addrlen);
+				sa->sockaddr_len = ai_ptr->ai_addrlen;
+				sa->sockaddr_time = now;
+				sa->fd = fd;
+
+				zend_hash_str_update_ptr(&resolver_cache, hostport, hostport_len, sa);
+				break;
+			}
+		}
+	}
+
+	freeaddrinfo(ai_list);
+	efree(hostport);
+	return sa;
+}
+/* }}} */
+
+static int php_pinba_init_socket(pinba_collector *collectors, int n_collectors) /* {{{ */
+{
 	int i;
-	int fd;
-	int status;
 	int n_fds;
 
 	if (n_collectors == 0) {
@@ -483,46 +540,11 @@ static int php_pinba_init_socket(pinba_collector *collectors, int n_collectors)
 	for (i = 0; i < n_collectors; i++) {
 		pinba_collector *collector = &collectors[i];
 
-		memset(&ai_hints, 0, sizeof(ai_hints));
-		ai_hints.ai_flags     = 0;
-#ifdef AI_ADDRCONFIG
-		ai_hints.ai_flags    |= AI_ADDRCONFIG;
-#endif
-		ai_hints.ai_family    = AF_UNSPEC;
-		ai_hints.ai_socktype  = SOCK_DGRAM;
-		ai_hints.ai_addr      = NULL;
-		ai_hints.ai_canonname = NULL;
-		ai_hints.ai_next      = NULL;
-
-		ai_list = NULL;
-		status = getaddrinfo(collector->host, collector->port, &ai_hints, &ai_list);
-		fd = -1; /* need to put -1 into collector->fd */
-		if (status != 0) {
-			php_error_docref(NULL, E_WARNING, "failed to resolve Pinba server hostname '%s': %s", collector->host, gai_strerror(status));
-		}
-		else {
-			for (ai_ptr = ai_list; ai_ptr != NULL; ai_ptr = ai_ptr->ai_next) {
-				fd = socket(ai_ptr->ai_family, ai_ptr->ai_socktype, ai_ptr->ai_protocol);
-				if (fd >= 0) {
-					break;
-				}
-			}
-		}
-
-		if (collector->fd >= 0) {
-			close(collector->fd);
-		}
-		collector->fd = fd;
-
-		if (fd < 0) {
+		pinba_sockaddr *sa = php_pinba_connect_socket(collector->host, collector->port);
+		if (!sa) {
 			continue; /* skip this one in case others are good */
 		}
 
-		memcpy(&collector->sockaddr, ai_ptr->ai_addr, ai_ptr->ai_addrlen);
-		collector->sockaddr_len = ai_ptr->ai_addrlen;
-
-		freeaddrinfo(ai_list);
-
 		n_fds++;
 	}
 
@@ -662,7 +684,7 @@ static inline Pinba__Request *php_create_pinba_packet(pinba_client_t *client, co
 			struct mallinfo info;
 
 			info = mallinfo();
-			request->memory_footprint = info.arena + info.hblkhd;
+			request->memory_footprint = info.arena;// + info.hblkhd;
 		}
 #else
 		request->memory_footprint = 0;
@@ -936,11 +958,12 @@ static inline int php_pinba_req_data_send(pinba_client_t *client, const char *cu
 		for (i = 0; i < n_collectors; i++) {
 			pinba_collector *collector = &collectors[i];
 
-			if (collector->fd < 0) {
-				continue;
+			pinba_sockaddr *sa = php_pinba_connect_socket(collector->host, collector->port);
+			if (!sa) {
+				continue; /* skip this one in case others are good */
 			}
 
-			sent = sendto(collector->fd, data, data_len, 0, (struct sockaddr *) &collector->sockaddr, collector->sockaddr_len);
+			sent = sendto(sa->fd, data, data_len, 0, (struct sockaddr *) &sa->sockaddr, sa->sockaddr_len);
 			if (sent < data_len) {
 				php_error_docref(NULL, E_WARNING, "failed to send data to Pinba server: %s", strerror(errno));
 				ret = FAILURE;
@@ -1176,9 +1199,8 @@ static void php_pinba_get_timer_info(pinba_timer_t *t, zval *info, struct timeva
 }
 /* }}} */
 
-static void pinba_client_free_storage(zend_object *object) /* {{{ */
+static void pinba_client_object_dtor(zend_object *object) /* {{{ */
 {
-	int i;
 	pinba_client_t *client = (pinba_client_t *) php_pinba_client_object(object);
 
 	if (!client->data_sent && (client->flags & PINBA_AUTO_FLUSH) != 0) {
@@ -1186,6 +1208,13 @@ static void pinba_client_free_storage(zend_object *object) /* {{{ */
 			php_pinba_req_data_send(client, NULL, client->flags);
 		}
 	}
+}
+/* }}} */
+
+static void pinba_client_free_storage(zend_object *object) /* {{{ */
+{
+	int i;
+	pinba_client_t *client = (pinba_client_t *) php_pinba_client_object(object);
 
 	zend_object_std_dtor(&client->std);
 
@@ -1241,7 +1270,7 @@ zend_object *pinba_client_new(zend_class_entry *ce) /* {{{ */
    Start user timer */
 static PHP_FUNCTION(pinba_timer_start)
 {
-	HashTable *tags_array;
+	zval *tags_array;
 	zval *data = NULL;
 	pinba_timer_t *t = NULL;
 	pinba_timer_tag_t **tags;
@@ -1255,11 +1284,14 @@ static PHP_FUNCTION(pinba_timer_start)
 		RETURN_FALSE;
 	}
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "h|al", &tags_array, &data, &hit_count) != SUCCESS) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 3)
+		Z_PARAM_ARRAY_EX(tags_array, 0, 1)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_ZVAL(data)
+		Z_PARAM_LONG(hit_count)
+	ZEND_PARSE_PARAMETERS_END_EX(RETURN_FALSE);
 
-	tags_num = zend_hash_num_elements(tags_array);
+	tags_num = zend_hash_num_elements(Z_ARRVAL_P(tags_array));
 
 	if (!tags_num) {
 		php_error_docref(NULL, E_WARNING, "tags array cannot be empty");
@@ -1271,7 +1303,7 @@ static PHP_FUNCTION(pinba_timer_start)
 		RETURN_FALSE;
 	}
 
-	if (php_pinba_array_to_tags(tags_array, &tags) != SUCCESS) {
+	if (php_pinba_array_to_tags(Z_ARRVAL_P(tags_array), &tags) != SUCCESS) {
 		RETURN_FALSE;
 	}
 
@@ -1292,16 +1324,20 @@ static PHP_FUNCTION(pinba_timer_start)
 		timeval_cvt(&t->tmp_ru_stime, &u.ru_stime);
 	}
 	/* refcount++ so that the timer is shut down only on request finish if not stopped manually */
+#if PHP_VERSION_ID < 70300
 	GC_REFCOUNT(rsrc)++;
+#else
+	GC_ADDREF(rsrc);
+#endif
 	RETURN_RES(rsrc);
 }
 /* }}} */
 
-/* {{{ proto resource pinba_timer_add(array tags, float value[, array data])
+/* {{{ proto resource pinba_timer_add(array tags, float value[[, array data,], int hit_count])
    Create user timer with a value */
 static PHP_FUNCTION(pinba_timer_add)
 {
-	HashTable *tags_array;
+	zval *tags_array;
 	zval *data = NULL;
 	pinba_timer_t *t = NULL;
 	pinba_timer_tag_t **tags;
@@ -1309,24 +1345,34 @@ static PHP_FUNCTION(pinba_timer_add)
 	double value;
 	unsigned long time_l;
 	zend_resource *rsrc;
+	long hit_count = 1;
 
 	if (PINBA_G(timers_stopped)) {
 		php_error_docref(NULL, E_WARNING, "all timers have already been stopped");
 		RETURN_FALSE;
 	}
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "hd|a", &tags_array, &value, &data) != SUCCESS) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(2, 4)
+		Z_PARAM_ARRAY_EX(tags_array, 0, 1)
+		Z_PARAM_DOUBLE(value)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_ZVAL(data)
+		Z_PARAM_LONG(hit_count)
+	ZEND_PARSE_PARAMETERS_END_EX(RETURN_FALSE);
 
-	tags_num = zend_hash_num_elements(tags_array);
+	tags_num = zend_hash_num_elements(Z_ARRVAL_P(tags_array));
 
 	if (!tags_num) {
 		php_error_docref(NULL, E_WARNING, "tags array cannot be empty");
 		RETURN_FALSE;
 	}
 
-	if (php_pinba_array_to_tags(tags_array, &tags) != SUCCESS) {
+	if (hit_count <= 0) {
+		php_error_docref(NULL, E_WARNING, "hit_count must be greater than 0 (%ld was passed)", hit_count);
+		RETURN_FALSE;
+	}
+
+	if (php_pinba_array_to_tags(Z_ARRVAL_P(tags_array), &tags) != SUCCESS) {
 		RETURN_FALSE;
 	}
 
@@ -1337,12 +1383,12 @@ static PHP_FUNCTION(pinba_timer_add)
 
 	t = php_pinba_timer_ctor(tags, tags_num);
 
-	if (data) {
+	if (data && zend_hash_num_elements(Z_ARRVAL_P(data)) > 0) {
 		ZVAL_DUP(&t->data, data);
 	}
 
 	t->started = 0;
-	t->hit_count = 1;
+	t->hit_count = hit_count;
 	time_l = (unsigned long)(value * 1000000.0);
 	t->value.tv_sec = time_l / 1000000;
 	t->value.tv_usec = time_l % 1000000;
@@ -1351,7 +1397,11 @@ static PHP_FUNCTION(pinba_timer_add)
 	t->rsrc_id = rsrc->handle;
 
 	/* refcount++ so that the timer is shut down only on request finish if not stopped manually */
+#if PHP_VERSION_ID < 70300
 	GC_REFCOUNT(rsrc)++;
+#else
+	GC_ADDREF(rsrc);
+#endif
 	RETURN_RES(rsrc);
 }
 /* }}} */
@@ -1474,7 +1524,7 @@ static PHP_FUNCTION(pinba_timer_data_replace)
    Merge timer data with new data */
 static PHP_FUNCTION(pinba_timer_tags_merge)
 {
-	HashTable *tags;
+	zval *tags;
 	zval *timer;
 	pinba_timer_t *t;
 	pinba_timer_tag_t **new_tags;
@@ -1485,19 +1535,20 @@ static PHP_FUNCTION(pinba_timer_tags_merge)
 		RETURN_FALSE;
 	}
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rh", &timer, &tags) != SUCCESS) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(2, 2)
+		Z_PARAM_RESOURCE(timer)
+		Z_PARAM_ARRAY_EX(tags, 0, 1)
+	ZEND_PARSE_PARAMETERS_END_EX(RETURN_FALSE);
 
 	PHP_ZVAL_TO_TIMER(timer, t);
 
-	tags_num = zend_hash_num_elements(tags);
+	tags_num = zend_hash_num_elements(Z_ARRVAL_P(tags));
 
 	if (!tags_num) {
 		RETURN_TRUE;
 	}
 
-	if (php_pinba_array_to_tags(tags, &new_tags) != SUCCESS) {
+	if (php_pinba_array_to_tags(Z_ARRVAL_P(tags), &new_tags) != SUCCESS) {
 		RETURN_FALSE;
 	}
 
@@ -1539,7 +1590,7 @@ static PHP_FUNCTION(pinba_timer_tags_merge)
    Replace timer data with new one */
 static PHP_FUNCTION(pinba_timer_tags_replace)
 {
-	HashTable *tags;
+	zval *tags;
 	zval *timer;
 	pinba_timer_t *t;
 	pinba_timer_tag_t **new_tags;
@@ -1550,20 +1601,21 @@ static PHP_FUNCTION(pinba_timer_tags_replace)
 		RETURN_FALSE;
 	}
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rh", &timer, &tags) != SUCCESS) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(2, 2)
+		Z_PARAM_RESOURCE(timer)
+		Z_PARAM_ARRAY_EX(tags, 0, 1)
+	ZEND_PARSE_PARAMETERS_END_EX(RETURN_FALSE);
 
 	PHP_ZVAL_TO_TIMER(timer, t);
 
-	tags_num = zend_hash_num_elements(tags);
+	tags_num = zend_hash_num_elements(Z_ARRVAL_P(tags));
 
 	if (!tags_num) {
 		php_error_docref(NULL, E_WARNING, "tags array cannot be empty");
 		RETURN_TRUE;
 	}
 
-	if (php_pinba_array_to_tags(tags, &new_tags) != SUCCESS) {
+	if (php_pinba_array_to_tags(Z_ARRVAL_P(tags), &new_tags) != SUCCESS) {
 		RETURN_FALSE;
 	}
 
@@ -1808,7 +1860,11 @@ static PHP_FUNCTION(pinba_timers_get)
 				continue;
 			}
 			/* refcount++ */
+#if PHP_VERSION_ID < 70300
 			GC_REFCOUNT(rsrc)++;
+#else
+			GC_ADDREF(rsrc);
+#endif
 			add_next_index_resource(return_value, rsrc);
 		}
 	}
@@ -2013,21 +2069,24 @@ static PHP_FUNCTION(pinba_tags_get)
     */
 static PHP_METHOD(PinbaClient, __construct)
 {
-	HashTable *servers;
+	zval *servers;
 	zval *tmp;
 	pinba_collector *new_collector;
 	pinba_client_t *client;
 	long flags = 0;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "h|l", &servers, &flags) != SUCCESS) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 2)
+		Z_PARAM_ARRAY_EX(servers, 0, 1)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_LONG(flags)
+	ZEND_PARSE_PARAMETERS_END_EX(RETURN_FALSE);
+
 	client = Z_PINBACLIENT_P(getThis());
 	client->flags = flags;
 
-	for (zend_hash_internal_pointer_reset(servers);
-		 (tmp = zend_hash_get_current_data(servers)) != NULL;
-		 zend_hash_move_forward(servers)) {
+	for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(servers));
+		 (tmp = zend_hash_get_current_data(Z_ARRVAL_P(servers))) != NULL;
+		 zend_hash_move_forward(Z_ARRVAL_P(servers))) {
 		char *host, *port, *address_copy;
 		zend_string *str = zval_get_string(tmp);
 
@@ -2047,7 +2106,6 @@ static PHP_METHOD(PinbaClient, __construct)
 
 		new_collector->host = strdup(host);
 		new_collector->port = (port == NULL) ? strdup(PINBA_COLLECTOR_DEFAULT_PORT) : strdup(port);
-		new_collector->fd = -1; /* set invalid fd */
 		efree(address_copy);
 	}
 }
@@ -2107,24 +2165,25 @@ SET_METHOD_NUM(setRequestTime, request_time, double, "d");
     */
 static PHP_METHOD(PinbaClient, setRusage)
 {
-	HashTable *rusage;
+	zval *rusage;
 	pinba_client_t *client;
 	zval *tmp;
 	int i;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "h", &rusage) != SUCCESS) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 1)
+		Z_PARAM_ARRAY_EX(rusage, 0, 1)
+	ZEND_PARSE_PARAMETERS_END_EX(RETURN_FALSE);
+
 	client = Z_PINBACLIENT_P(getThis());
 
-	if (zend_hash_num_elements(rusage) != 2) {
+	if (zend_hash_num_elements(Z_ARRVAL_P(rusage)) != 2) {
 		php_error_docref(NULL, E_WARNING, "rusage array must contain exactly 2 elements");
 		RETURN_FALSE;
 	}
 
-	for (zend_hash_internal_pointer_reset(rusage), i = 0;
-		 ((tmp = zend_hash_get_current_data(rusage)) != NULL) && i < 2;
-		 zend_hash_move_forward(rusage), i++) {
+	for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(rusage)), i = 0;
+		 ((tmp = zend_hash_get_current_data(Z_ARRVAL_P(rusage))) != NULL) && i < 2;
+		 zend_hash_move_forward(Z_ARRVAL_P(rusage)), i++) {
 
 		client->rusage[i] = zval_get_double(tmp);
 	}
@@ -2157,19 +2216,24 @@ static void php_pinba_client_timer_add_set(INTERNAL_FUNCTION_PARAMETERS, int add
 	pinba_client_t *client;
 	long hit_count = 1;
 	double value, ru_utime = 0, ru_stime = 0;
-	HashTable *tags, *rusage = NULL;
+	zval *tags, *rusage = NULL;
 	char *hashed_tags;
 	size_t hashed_tags_len, i, tags_num;
 	zval *tmp;
 	pinba_timer_t *timer;
 	pinba_timer_tag_t **new_tags;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "hd|hl", &tags, &value, &rusage, &hit_count) != SUCCESS) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(2, 4)
+		Z_PARAM_ARRAY_EX(tags, 0, 1)
+		Z_PARAM_DOUBLE(value)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_ARRAY_EX(rusage, 0, 1)
+		Z_PARAM_LONG(hit_count)
+	ZEND_PARSE_PARAMETERS_END_EX(RETURN_FALSE);
+
 	client = Z_PINBACLIENT_P(getThis());
 
-	tags_num = zend_hash_num_elements(tags);
+	tags_num = zend_hash_num_elements(Z_ARRVAL_P(tags));
 	if (!tags_num) {
 		php_error_docref(NULL, E_WARNING, "timer tags array cannot be empty");
 		RETURN_FALSE;
@@ -2185,15 +2249,15 @@ static void php_pinba_client_timer_add_set(INTERNAL_FUNCTION_PARAMETERS, int add
 		RETURN_FALSE;
 	}
 
-	if (rusage && zend_hash_num_elements(rusage) != 2) {
+	if (rusage && zend_hash_num_elements(Z_ARRVAL_P(rusage)) != 2) {
 		php_error_docref(NULL, E_WARNING, "rusage array must contain exactly 2 elements");
 		RETURN_FALSE;
 	}
 
 	if (rusage) {
-		for (zend_hash_internal_pointer_reset(rusage), i = 0;
-				((tmp = zend_hash_get_current_data(rusage)) != NULL) && i < 2;
-				zend_hash_move_forward(rusage), i++) {
+		for (zend_hash_internal_pointer_reset(Z_ARRVAL_P(rusage)), i = 0;
+				((tmp = zend_hash_get_current_data(Z_ARRVAL_P(rusage))) != NULL) && i < 2;
+				zend_hash_move_forward(Z_ARRVAL_P(rusage)), i++) {
 
 			if (i == 0) {
 				ru_utime = zval_get_double(tmp);
@@ -2203,7 +2267,7 @@ static void php_pinba_client_timer_add_set(INTERNAL_FUNCTION_PARAMETERS, int add
 		}
 	}
 
-	if (php_pinba_array_to_tags(tags, &new_tags) != SUCCESS) {
+	if (php_pinba_array_to_tags(Z_ARRVAL_P(tags), &new_tags) != SUCCESS) {
 		RETURN_FALSE;
 	}
 
@@ -2516,6 +2580,15 @@ zend_function_entry pinba_client_methods[] = {
 };
 /* }}} */
 
+static void php_pinba_sa_dtor(zval *zv) /* {{{ */
+{
+	pinba_sockaddr *sa = Z_PTR_P(zv);
+	if (sa->fd >= 0) {
+		close(sa->fd);
+	}
+	free(sa);
+}
+/* }}} */
 
 static PHP_INI_MH(OnUpdateCollectorAddress) /* {{{ */
 {
@@ -2551,7 +2624,6 @@ static PHP_INI_MH(OnUpdateCollectorAddress) /* {{{ */
 		}
 		new_collector->host = strdup(new_node);
 		new_collector->port = (new_service == NULL) ? strdup(PINBA_COLLECTOR_DEFAULT_PORT) : strdup(new_service);
-		new_collector->fd = -1; /* set invalid fd */
 	}
 
 	free(copy);
@@ -2565,6 +2637,7 @@ static PHP_INI_MH(OnUpdateCollectorAddress) /* {{{ */
  */
 PHP_INI_BEGIN()
     STD_PHP_INI_ENTRY("pinba.server", NULL, PHP_INI_ALL, OnUpdateCollectorAddress, collector_address, zend_pinba_globals, pinba_globals)
+    STD_PHP_INI_ENTRY("pinba.resolve_interval", "60", PHP_INI_ALL, OnUpdateLongGEZero, resolve_interval, zend_pinba_globals, pinba_globals)
     STD_PHP_INI_ENTRY("pinba.enabled", "0", PHP_INI_ALL, OnUpdateBool, enabled, zend_pinba_globals, pinba_globals)
     STD_PHP_INI_ENTRY("pinba.auto_flush", "1", PHP_INI_ALL, OnUpdateBool, auto_flush, zend_pinba_globals, pinba_globals)
 PHP_INI_END()
@@ -2609,10 +2682,12 @@ static PHP_MINIT_FUNCTION(pinba)
 	pinba_client_ce->create_object = pinba_client_new;
 
 	memcpy(&pinba_client_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
-	pinba_client_handlers.dtor_obj = zend_objects_destroy_object;
+	pinba_client_handlers.dtor_obj = pinba_client_object_dtor;
 	pinba_client_handlers.free_obj = pinba_client_free_storage;
 	pinba_client_handlers.clone_obj = NULL;
 	pinba_client_handlers.offset = XtOffsetOf(pinba_client_t, std);
+
+	zend_hash_init(&resolver_cache, 10, NULL, php_pinba_sa_dtor, 1);
 	return SUCCESS;
 }
 /* }}} */
@@ -2624,6 +2699,8 @@ static PHP_MSHUTDOWN_FUNCTION(pinba)
 	UNREGISTER_INI_ENTRIES();
 
 	php_pinba_cleanup_collectors(PINBA_G(collectors), &PINBA_G(n_collectors));
+
+	zend_hash_destroy(&resolver_cache);
 	return SUCCESS;
 }
 /* }}} */
